from typing import Sequence

from pydantic import BaseModel

from .base import Widget


class Checkable(BaseModel):
    label: str
    name: str
    value: str
    token: str
    checked: bool

    @property
    def id(self) -> str:
        id = f"{self.name}-{self.value}-{self.token}"
        return id.replace(".", "-").replace("_", "-")


class ChecklistWidget(Widget[Sequence[Checkable]]):
    def __init__(
        self,
        name: str,
        *,
        title: str | None,
        value: Sequence[Checkable],
        token: str,
        removable: bool,
    ) -> None:
        super().__init__(
            name, value=value, token=token, title=title, removable=removable
        )

    def get_template(self) -> str:
        return "pydantic_form.Checklist"



from .base import Widget


class BooleanWidget(Widget[bool]):
    def __init__(
        self,
        name: str,
        *,
        title: str | None,
        value: bool = False,
        removable: bool = False,
        token: str,
    ) -> None:
        super().__init__(
            name, title=title, value=value, removable=removable, token=token
        )

    def get_template(self) -> str:
        return "pydantic_form.Boolean"



import secrets
from typing import TYPE_CHECKING, Any, Callable, Coroutine

from fastapi import Request

if TYPE_CHECKING:
    from fastlife.configurator.registry import Registry  # coverage: ignore


class CSRFAttack(Exception):
    ...


def create_csrf_token() -> str:
    return secrets.token_urlsafe(5)


def check_csrf(registry: "Registry") -> Callable[[Request], Coroutine[Any, Any, bool]]:
    async def check_csrf(request: Request) -> bool:
        if (
            request.method != "POST"
            or request.headers.get("content-type")
            != "application/x-www-form-urlencoded"
        ):
            return True

        cookie = request.cookies.get(registry.settings.csrf_token_name)
        if not cookie:
            raise CSRFAttack("CSRF token did not match")

        form_data = await request.form()
        value = form_data.get(registry.settings.csrf_token_name)
        if value != cookie:
            raise CSRFAttack("CSRF token did not match")

        return True

    return check_csrf



import abc
from typing import Any, Mapping, Optional, Type

from fastapi import Request
from markupsafe import Markup
from pydantic.fields import FieldInfo


class AbstractTemplateRenderer(abc.ABC):
    route_prefix: str
    """Used to buid pydantic form"""

    @abc.abstractmethod
    def render_template(
        self,
        template: str,
        *,
        globals: Optional[Mapping[str, Any]] = None,
        **params: Any,
    ) -> str:
        """
        Render the given template with the given params.

        While rendering templates, the globals parameter is keps by the instantiated
        renderer and sent to every rendering made by the request.
        This is used by the pydantic form method that will render other templates
        for the request.
        """

    @abc.abstractmethod
    def pydantic_form(
        self,
        model: Type[Any],
        form_data: Mapping[str, Any] | None = None,
        name: str | None = None,
        token: str | None = None,
        removable: bool = False,
        field: FieldInfo | None = None,
    ) -> Markup:
        ...


class AbstractTemplateRendererFactory(abc.ABC):
    @abc.abstractmethod
    def __call__(self, request: Request) -> AbstractTemplateRenderer:
        ...



from typing import Optional, Sequence, Tuple

from .base import Widget


class DropDownWidget(Widget[str]):
    def __init__(
        self,
        name: str,
        *,
        title: Optional[str],
        value: Optional[str] = None,
        options: Sequence[Tuple[str, str]] | Sequence[str],
        removable: bool = False,
        token: Optional[str] = None,
        hint: Optional[str] = None,
    ) -> None:
        super().__init__(
            name, value=value, title=title, token=token, removable=removable
        )
        self.options: list[dict[str, str]] = []
        for opt in options:
            if isinstance(opt, tuple):
                self.options.append({"value": opt[0], "text": opt[1]})
            else:
                self.options.append({"value": opt, "text": opt})
        self.hint = hint

    def get_template(self) -> str:
        return "pydantic_form.Dropdown"



"""
The configurator is here to register routes in a fastapi app,
with dependency injection.
"""
import importlib
import inspect
import logging
from enum import Enum
from pathlib import Path
from types import ModuleType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Coroutine,
    List,
    Optional,
    Self,
    Type,
    Union,
)

import venusian  # type: ignore
from fastapi import Depends, FastAPI, Response
from fastapi.params import Depends as DependsType
from fastapi.staticfiles import StaticFiles

from fastlife.configurator.base import AbstractMiddleware
from fastlife.security.csrf import check_csrf

from .settings import Settings

if TYPE_CHECKING:
    from .registry import AppRegistry  # coverage: ignore

log = logging.getLogger(__name__)
VENUSIAN_CATEGORY = "fastlife"


class Configurator:
    registry: "AppRegistry"

    def __init__(self, settings: Settings) -> None:
        from .registry import initialize_registry  # XXX circular import

        self.registry = initialize_registry(settings)
        self._app = FastAPI(
            dependencies=[Depends(check_csrf(self.registry))],
            docs_url=None,
            redoc_url=None,
        )
        self.scanner = venusian.Scanner(fastlife=self)
        self.include("fastlife.views")
        self.include("fastlife.session")

    def get_app(self) -> FastAPI:
        return self._app

    def include(self, module: str | ModuleType) -> "Configurator":
        if isinstance(module, str):
            package = None
            if module.startswith("."):
                caller_module = inspect.getmodule(inspect.stack()[1][0])
                package = caller_module.__name__ if caller_module else "__main__"

            module = importlib.import_module(module, package)
        self.scanner.scan(module, categories=[VENUSIAN_CATEGORY])  # type: ignore
        return self

    def add_middleware(
        self, middleware_class: Type[AbstractMiddleware], **options: Any
    ) -> Self:
        self._app.add_middleware(middleware_class, **options)
        return self

    def add_route(
        self,
        name: str,
        path: str,
        endpoint: Callable[..., Coroutine[Any, Any, Response]],
        *,
        permission: str | None = None,
        status_code: int | None = None,
        tags: List[Union[str, Enum]] | None = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_description: str = "Successful Response",
        # responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        methods: Optional[List[str]] = None,
        # operation_id: Optional[str] = None,
        # response_model: Any = Default(None),
        # response_model_include: Optional[IncEx] = None,
        # response_model_exclude: Optional[IncEx] = None,
        # response_model_by_alias: bool = True,
        # response_model_exclude_unset: bool = False,
        # response_model_exclude_defaults: bool = False,
        # response_model_exclude_none: bool = False,
        # include_in_schema: bool = True,
        # response_class: Union[Type[Response], DefaultPlaceholder] = Default(
        #     HTMLResponse
        # ),
        # openapi_extra: Optional[Dict[str, Any]] = None,
        # generate_unique_id_function: Callable[[APIRoute], str] = Default(
        #     generate_unique_id
        # ),
    ) -> "Configurator":
        dependencies: List[DependsType] = []
        if permission:
            dependencies.append(Depends(self.registry.check_permission(permission)))

        self._app.add_api_route(
            path,
            endpoint,
            # response_model=response_model,
            status_code=status_code,
            tags=tags,
            dependencies=dependencies,
            summary=summary,
            description=description,
            response_description=response_description,
            # responses=responses,
            deprecated=deprecated,
            methods=methods,
            # operation_id=operation_id,
            # response_model_include=response_model_include,
            # response_model_exclude=response_model_exclude,
            # response_model_by_alias=response_model_by_alias,
            # response_model_exclude_unset=response_model_exclude_unset,
            # response_model_exclude_defaults=response_model_exclude_defaults,
            # response_model_exclude_none=response_model_exclude_none,
            # include_in_schema=include_in_schema,
            # response_class=response_class,
            name=name,
            # openapi_extra=openapi_extra,
            # generate_unique_id_function=generate_unique_id_function,
        )
        return self

    def add_static_route(
        self, route_path: str, directory: Path, name: str = "static"
    ) -> "Configurator":
        """Mount a directory to an http endpoint."""
        self._app.mount(route_path, StaticFiles(directory=directory), name=name)
        return self

    def add_exception_handler(
        self, status_code_or_exc: int | Type[Exception], handler: Any
    ) -> "Configurator":
        self._app.add_exception_handler(status_code_or_exc, handler)
        return self


def configure(
    wrapped: Callable[[Configurator], None]
) -> Callable[[Configurator], None]:
    """Decorator used to attach route in a submodule while using the confirator.scan"""

    def callback(
        scanner: venusian.Scanner, name: str, ob: Callable[[venusian.Scanner], None]
    ) -> None:
        if hasattr(scanner, "fastlife"):
            ob(scanner.fastlife)  # type: ignore

    venusian.attach(wrapped, callback, category=VENUSIAN_CATEGORY)  # type: ignore
    return wrapped



from typing import TYPE_CHECKING, Annotated

from fastapi import Depends

from fastlife.security.policy import CheckPermission
from fastlife.shared_utils.resolver import resolve

if TYPE_CHECKING:
    from fastlife.templating.renderer import (  # coverage: ignore
        AbstractTemplateRendererFactory,  # coverage: ignore
    )  # coverage: ignore

from .settings import Settings


class AppRegistry:
    settings: Settings
    renderer: "AbstractTemplateRendererFactory"
    check_permission: CheckPermission

    def __init__(self, settings: Settings) -> None:
        # Abtract class resolved for dependency injection
        TemplateRenderer = resolve(settings.template_renderer_class)

        self.settings = settings
        self.renderer = TemplateRenderer(settings)
        self.check_permission = resolve(settings.check_permission)


DEFAULT_REGISTRY: AppRegistry = None  # type: ignore


def initialize_registry(settings: Settings) -> AppRegistry:
    global DEFAULT_REGISTRY
    if DEFAULT_REGISTRY is not None:
        raise ValueError("Registry is already set")
    AppRegistryCls = resolve(settings.registry_class)
    DEFAULT_REGISTRY = AppRegistryCls(settings)  # type: ignore
    return DEFAULT_REGISTRY


def cleanup_registry() -> None:
    """ "Method to cleanup the registry, used for testing"""
    global DEFAULT_REGISTRY
    DEFAULT_REGISTRY = None  # type: ignore


Registry = Annotated[AppRegistry, Depends(lambda: DEFAULT_REGISTRY)]



from .binding import Template, template
from .renderer import AbstractTemplateRendererFactory, JinjaxTemplateRenderer

__all__ = [
    "AbstractTemplateRendererFactory",
    "JinjaxTemplateRenderer",
    "Template",
    "template",
]



from typing import Optional

from .base import Widget


class TextWidget(Widget[str]):
    def __init__(
        self,
        name: str,
        *,
        title: Optional[str],
        aria_label: Optional[str] = None,
        placeholder: Optional[str] = None,
        removable: bool = False,
        value: str = "",
        token: Optional[str] = None,
        hint: Optional[str] = None,
        input_type: str = "text"
    ) -> None:
        super().__init__(
            name,
            value=value,
            title=title,
            aria_label=aria_label,
            token=token,
            removable=removable,
        )
        self.placeholder = placeholder or ""
        self.hint = hint
        self.input_type = input_type

    def get_template(self) -> str:
        return "pydantic_form.Text"



import secrets
from collections.abc import MutableSequence, Sequence
from decimal import Decimal
from enum import Enum
from types import NoneType
from typing import Any, Literal, Mapping, Optional, Type, cast, get_origin
from uuid import UUID

from markupsafe import Markup
from pydantic import BaseModel, EmailStr, SecretStr, ValidationError
from pydantic.fields import FieldInfo

from fastlife.shared_utils.infer import is_complex_type, is_union
from fastlife.templating.renderer.abstract import AbstractTemplateRenderer
from fastlife.templating.renderer.widgets.boolean import BooleanWidget
from fastlife.templating.renderer.widgets.checklist import Checkable, ChecklistWidget
from fastlife.templating.renderer.widgets.dropdown import DropDownWidget
from fastlife.templating.renderer.widgets.hidden import HiddenWidget
from fastlife.templating.renderer.widgets.sequence import SequenceWidget

from .base import Widget, get_title
from .model import ModelWidget
from .text import TextWidget
from .union import UnionWidget


class WidgetFactory:
    def __init__(self, renderer: AbstractTemplateRenderer, token: Optional[str] = None):
        self.renderer = renderer
        self.token = token or secrets.token_urlsafe(4).replace("_", "-")

    def get_markup(
        self,
        base: Type[Any],
        form_data: Mapping[str, Any],
        *,
        prefix: str,
        removable: bool,
        field: FieldInfo | None = None,
    ) -> Markup:
        return self.get_widget(
            base, form_data, prefix=prefix, removable=removable, field=field
        ).to_html(self.renderer)

    def get_widget(
        self,
        base: Type[Any],
        form_data: Mapping[str, Any],
        *,
        prefix: str,
        removable: bool,
        field: FieldInfo | None = None,
    ) -> Widget[Any]:
        return self.build(
            base,
            value=form_data.get(prefix, {}),
            name=prefix,
            removable=removable,
            field=field,
        )

    def build(
        self,
        typ: Type[Any],
        *,
        name: str = "",
        value: Any,
        removable: bool,
        field: FieldInfo | None = None,
    ) -> Widget[Any]:
        if field and field.metadata:
            for widget in field.metadata:
                if issubclass(widget, Widget):
                    return cast(
                        Widget[Any],
                        widget(
                            name,
                            value=value,
                            removable=removable,
                            title=field.title if field else "",
                            aria_label=field.description if field else None,
                            token=self.token,
                        ),
                    )

        type_origin = get_origin(typ)
        if type_origin:
            if is_union(typ):
                return self.build_union(name, typ, field, value, removable)

            if (
                type_origin is Sequence
                or type_origin is MutableSequence
                or type_origin is list
            ):
                return self.build_sequence(name, typ, field, value, removable)

            if type_origin is Literal:
                return self.build_literal(name, typ, field, value, removable)

            if type_origin is set:
                return self.build_set(name, typ, field, value, removable)

        if issubclass(typ, Enum):  # if it raises here, the type_origin is unknown
            return self.build_enum(name, typ, field, value, removable)

        if issubclass(typ, BaseModel):  # if it raises here, the type_origin is unknown
            return self.build_model(name, typ, field, value or {}, removable)

        if issubclass(typ, bool):
            return self.build_boolean(name, typ, field, value or False, removable)

        if issubclass(typ, EmailStr):  # type: ignore
            return self.build_emailtype(name, typ, field, value or "", removable)

        if issubclass(typ, SecretStr):
            return self.build_secretstr(name, typ, field, value or "", removable)

        if issubclass(typ, (int, str, float, Decimal, UUID)):
            return self.build_simpletype(name, typ, field, value or "", removable)

        raise NotImplementedError(f"{typ} not implemented")  # coverage: ignore

    def build_model(
        self,
        field_name: str,
        typ: Type[BaseModel],
        field: Optional[FieldInfo],
        value: Mapping[str, Any],
        removable: bool,
    ) -> Widget[Any]:
        ret: dict[str, Any] = {}
        for key, field in typ.model_fields.items():
            child_key = f"{field_name}.{key}" if field_name else key
            if field.exclude:
                continue
            if field.annotation is None:
                raise ValueError(  # coverage: ignore
                    f"Missing annotation for {field} in {child_key}"
                )
            ret[key] = self.build(
                field.annotation,
                name=child_key,
                field=field,
                value=value.get(key),
                removable=False,
            )
        return ModelWidget(
            field_name,
            value=list(ret.values()),
            removable=removable,
            title=get_title(typ),
            token=self.token,
        )

    def build_union(
        self,
        field_name: str,
        field_type: Type[Any],
        field: Optional[FieldInfo],
        value: Any,
        removable: bool,
    ) -> Widget[Any]:
        types: list[Type[Any]] = []
        # required = True
        for typ in field_type.__args__:  # type: ignore
            if typ is NoneType:
                # required = False
                continue
            types.append(typ)  # type: ignore

        if (
            not removable
            and len(types) == 1
            # if the optional type is a complex type,
            and not is_complex_type(types[0])
        ):
            return self.build(
                types[0], name=field_name, field=field, value=value, removable=False
            )
        child = None
        if value:
            for typ in types:
                try:
                    typ(**value)
                except ValidationError:
                    pass
                else:
                    child = self.build(
                        typ,
                        name=field_name,
                        field=field,
                        value=value,
                        removable=False,
                    )

        widget = UnionWidget(
            field_name,
            # we assume those types are BaseModel
            value=child,
            children_types=types,  # type: ignore
            title=field.title if field else "",
            token=self.token,
            removable=removable,
        )

        return widget

    def build_sequence(
        self,
        field_name: str,
        field_type: Type[Any],
        field: Optional[FieldInfo],
        value: Optional[Sequence[Any]],
        removable: bool,
    ) -> Widget[Any]:
        typ = field_type.__args__[0]  # type: ignore
        value = value or []
        items = [
            self.build(
                typ,  # type: ignore
                name=f"{field_name}.{idx}",
                value=v,
                field=field,
                removable=True,
            )
            for idx, v in enumerate(value)
        ]
        return SequenceWidget(
            field_name,
            hint=field.description if field else "",
            title=field.title if field else "",
            value=items,
            item_type=typ,  # type: ignore
            token=self.token,
            removable=removable,
        )

    def build_set(
        self,
        field_name: str,
        field_type: Type[Any],
        field: Optional[FieldInfo],
        value: Optional[Sequence[Any]],
        removable: bool,
    ) -> Widget[Any]:
        choice_wrapper = field_type.__args__[0]
        choices = []
        choice_wrapper_origin = get_origin(choice_wrapper)
        if choice_wrapper_origin:
            if choice_wrapper_origin is Literal:
                litchoice: list[str] = choice_wrapper.__args__  # type: ignore
                choices = [
                    Checkable(
                        label=c,
                        value=c,
                        checked=c in value if value else False,  # type: ignore
                        name=field_name,
                        token=self.token,
                    )
                    for c in litchoice
                ]

            else:
                raise NotImplementedError
        elif issubclass(choice_wrapper, Enum):
            choices = [
                Checkable(
                    label=e.value,
                    value=e.name,
                    checked=e.name in value if value else False,  # type: ignore
                    name=field_name,
                    token=self.token,
                )
                for e in choice_wrapper
            ]
        else:
            raise NotImplementedError

        return ChecklistWidget(
            field_name,
            title=field.title if field else "",
            token=self.token,
            value=choices,
            removable=removable,
        )

    def build_boolean(
        self,
        field_name: str,
        field_type: Type[Any],
        field: FieldInfo | None,
        value: bool,
        removable: bool,
    ) -> Widget[Any]:
        return BooleanWidget(
            field_name,
            removable=removable,
            title=field.title if field else "",
            token=self.token,
            value=value,
        )

    def build_emailtype(
        self,
        field_name: str,
        field_type: Type[Any],
        field: FieldInfo | None,
        value: str | int | float,
        removable: bool,
    ) -> Widget[Any]:
        return TextWidget(
            field_name,
            hint=field.description if field else "",
            input_type="email",
            placeholder=str(field.examples[0]) if field and field.examples else None,
            removable=removable,
            title=field.title if field else "",
            token=self.token,
            value=str(value),
        )

    def build_secretstr(
        self,
        field_name: str,
        field_type: Type[Any],
        field: FieldInfo | None,
        value: SecretStr | str,
        removable: bool,
    ) -> Widget[Any]:
        return TextWidget(
            field_name,
            hint=field.description if field else "",
            input_type="password",
            placeholder=str(field.examples[0]) if field and field.examples else None,
            removable=removable,
            title=field.title if field else "",
            token=self.token,
            value=value.get_secret_value() if isinstance(value, SecretStr) else value,
        )

    def build_literal(
        self,
        field_name: str,
        field_type: Type[Any],  # a literal actually
        field: FieldInfo | None,
        value: str | int | float,
        removable: bool,
    ) -> Widget[Any]:
        choices: list[str] = field_type.__args__  # type: ignore
        if len(choices) == 1:
            return HiddenWidget(
                field_name,
                value=choices[0],
                token=self.token,
            )
        return DropDownWidget(
            field_name,
            options=choices,
            removable=removable,
            title=field.title if field else "",
            token=self.token,
            value=str(value),
        )

    def build_enum(
        self,
        field_name: str,
        field_type: Type[Any],  # an enum subclass
        field: FieldInfo | None,
        value: str | int | float,
        removable: bool,
    ) -> Widget[Any]:
        options = [(item.name, item.value) for item in field_type]  # type: ignore
        return DropDownWidget(
            field_name,
            options=options,  # type: ignore
            removable=removable,
            title=field.title if field else "",
            token=self.token,
            value=str(value),
        )

    def build_simpletype(
        self,
        field_name: str,
        field_type: Type[Any],
        field: FieldInfo | None,
        value: str | int | float,
        removable: bool,
    ) -> Widget[Any]:
        return TextWidget(
            field_name,
            hint=field.description if field else None,
            placeholder=str(field.examples[0]) if field and field.examples else None,
            aria_label=field.description if field else None,
            removable=removable,
            title=field.title if field else "",
            token=self.token,
            value=str(value),
        )



from fastlife import Configurator, configure
from fastlife.shared_utils.resolver import resolve

from .middleware import SessionMiddleware


@configure
def includeme(config: Configurator) -> None:
    settings = config.registry.settings
    session_serializer = resolve(settings.session_serializer)
    if settings.session_secret_key:
        config.add_middleware(
            SessionMiddleware,
            cookie_name=settings.session_cookie_name,
            secret_key=settings.session_secret_key,
            duration=settings.session_duration,
            cookie_path=settings.session_cookie_path,
            cookie_same_site=settings.session_cookie_same_site,
            cookie_secure=settings.session_cookie_secure,
            cookie_domain=settings.session_cookie_domain,
            serializer=session_serializer,
        )



from .configurator import Configurator, configure
from .configurator.registry import Registry
from .request.form_data import model
from .templating import Template, template

__all__ = [
    # Config
    "configure",
    "Configurator",
    "template",
    "Template",
    "Registry",
    # Model
    "model",
]



from typing import (
    Annotated,
    Any,
    Callable,
    Mapping,
    MutableMapping,
    MutableSequence,
    Optional,
    Sequence,
    Type,
    TypeVar,
)

from fastapi import Depends, Request
from pydantic import BaseModel

from fastlife.configurator.registry import Registry


def unflatten_struct(
    flatten_input: Mapping[str, Any],
    unflattened_output: MutableMapping[str, Any] | MutableSequence[Any],
    level: int = 0,
    *,
    csrf_token_name: Optional[str] = None,
) -> Mapping[str, Any] | Sequence[Any]:
    # we sort to ensure that list index are ordered
    # formkeys = sorted(flatten_input.keys())
    for key in flatten_input:
        if csrf_token_name is not None and key == csrf_token_name:
            continue
        lkey, sep, rest = key.partition(".")

        if not sep:
            # this is a leaf
            if isinstance(unflattened_output, list):
                if not key.isdigit():
                    raise ValueError(f"{flatten_input}: Not a list")
                while int(key) != len(unflattened_output):
                    unflattened_output.append(None)
                unflattened_output.append(flatten_input[key])
            elif isinstance(unflattened_output, dict):
                unflattened_output[key] = flatten_input[key]
            else:
                raise TypeError(type(unflattened_output))
            continue

        child_is_list = rest.partition(".")[0].isdigit()
        if isinstance(unflattened_output, list):
            vkey = int(lkey)
            while len(unflattened_output) < vkey:
                unflattened_output.append(None)
            if len(unflattened_output) < vkey + 1:
                sub_child_is_list = rest.partition(".")[0].isdigit()
                unflattened_output.append([] if sub_child_is_list else {})

            unflatten_struct(
                {rest: flatten_input[key]},
                unflattened_output[vkey],
                level + 1,
            )

        elif isinstance(unflattened_output, dict):
            if lkey not in unflattened_output:
                unflattened_output[lkey] = [] if child_is_list else {}
            unflatten_struct(
                {rest: flatten_input[key]},
                unflattened_output[lkey],
                level + 1,
            )
        else:
            raise TypeError(type(unflattened_output))

    return unflattened_output


async def unflatten_mapping_form_data(
    request: Request, reg: Registry
) -> Mapping[str, Any]:
    form_data = await request.form()
    form_data_decode_list: MutableMapping[str, Any] = {}
    for key, val in form_data.multi_items():
        if key in form_data_decode_list:
            if not isinstance(form_data_decode_list, list):
                form_data_decode_list[key] = [form_data_decode_list[key]]
            form_data_decode_list[key].append(val)
        else:
            form_data_decode_list[key] = val

    ret = unflatten_struct(
        form_data_decode_list, {}, csrf_token_name=reg.settings.csrf_token_name
    )  # type: ignore
    return ret  # type: ignore


async def unflatten_sequence_form_data(
    request: Request, reg: Registry
) -> Sequence[str]:
    form_data = await request.form()
    # Could raise a value error !
    return unflatten_struct(
        form_data, [], csrf_token_name=reg.settings.csrf_token_name
    )  # type: ignore


MappingFormData = Annotated[Mapping[str, Any], Depends(unflatten_mapping_form_data)]
SequenceFormData = Annotated[Sequence[str], Depends(unflatten_sequence_form_data)]


T = TypeVar("T", bound=BaseModel)
"""Template type for form serialized model"""


def model(cls: Type[T], name: Optional[str] = None) -> Callable[[Mapping[str, Any]], T]:
    """
    Build a model, a class of type T based on Pydandic Base Model from a form payload.
    """

    def to_model(data: MappingFormData, registry: Registry) -> Optional[T]:
        if data:
            return cls(**data[name or registry.settings.form_data_model_prefix])
        return None

    return Depends(to_model)



from typing import Any, Callable, Coroutine

CheckPermissionHook = Callable[..., Coroutine[Any, Any, None]] | Callable[..., None]
CheckPermission = Callable[[str], CheckPermissionHook]


def check_permission(permission_name: str) -> CheckPermissionHook:
    """
    A closure that check that a user as the given username.

    This method has to be overriden using the setting check_permission
    to implement it.
    """

    def depencency_injection() -> None:
        ...

    return depencency_injection
